// SPDX-License-Identifier: AGPL-3.0-or-later
//! This module defines [HashMapWithDefault].

use std::{borrow::Borrow, collections::HashMap, hash::Hash, ops::Index};

/// A newtype wrapping a [HashMap] which allows
/// a default value to be defined.
/// When a key is not found in the [HashMap] the default value is returned, instead of [None].
pub struct HashMapWithDefault<K, V> {
    hash_map: HashMap<K, V>,
    default: V,
}

impl<K, V> HashMapWithDefault<K, V> {
    /// Create a new hash map with the given default value.
    pub fn new(default: V) -> Self {
        Self {
            hash_map: HashMap::new(),
            default,
        }
    }

    /// Get a reference to the inner [HashMap].
    pub fn get_ref(&self) -> &HashMap<K, V> {
        &self.hash_map
    }

    /// Get a mutable reference to the inner [HashMap].
    pub fn get_mut(&mut self) -> &mut HashMap<K, V> {
        &mut self.hash_map
    }

    /// Returns a reference to the default value.
    pub fn default(&self) -> &V {
        &self.default
    }

    /// Returns a mutable reference to the default value.
    pub fn mut_default(&mut self) -> &mut V {
        &mut self.default
    }
}

impl<K: Hash + Eq, V> HashMapWithDefault<K, V> {
    /// Lookup up the item with the given key in the hash table. If no value is found then the
    /// default is returned.
    pub fn get<Q: ?Sized + Hash + Eq>(&self, index: &Q) -> &V
    where
        K: Borrow<Q>,
    {
        self.hash_map.get(index).unwrap_or(&self.default)
    }
}

impl<Q: ?Sized + Hash + Eq, K: Borrow<Q> + Hash + Eq, V> Index<&Q> for HashMapWithDefault<K, V> {
    type Output = V;
    fn index(&self, index: &Q) -> &Self::Output {
        self.get(index)
    }
}
